from ..utils.logging import logger
from ..utils.config import Config
from ..utils.context import Context
import sys
import re
import json
from ..utils.common import delete_stack, remove_change_set, format_diff, get_yes_or_no, deploy_stack, create_stack,package


def destroy(config_path, root_path):
    config = Config.parse(config_path)
    config.setup_env(Context.get_args().env)
    services = config.stacks
    
    for service in services:
        if not re.search(Context.get_args().service, service.name):
            continue

                
        logger.error(f"❗❗Warning about to delete stack {service.name}❗❗")
        should_continue = get_yes_or_no(f"Are you sure that you want to delete: {service.name}?")

        if should_continue:
            delete_stack(service.name)


        



from typing import List, Optional
from pydantic import BaseModel
import yaml
import os
from .context import Context

class Stack(BaseModel):
    path: str
    name: str
    envs: Optional[List[str]] = None
    parameters: Optional[dict] = None

    @property
    def _path(self):
        return self.path.replace("$root", Context.get_root())
        
    @property
    def _yml(self):
        class SafeLoaderIgnoreUnknown(yaml.SafeLoader):
            def ignore_unknown(self, node):
                return None 

        SafeLoaderIgnoreUnknown.add_constructor(None, SafeLoaderIgnoreUnknown.ignore_unknown)

        return yaml.load(open(self._path).read(), Loader=SafeLoaderIgnoreUnknown)

    @property
    def resources(self):
        return self._yml.get("Resources", {})

class Enviroment(BaseModel):
    artifacts: str 
    name: str
    region: Optional[str] = None
    profile: Optional[str] = None



class Config(BaseModel):
    Stacks: List[Stack]
    Environments: List[Enviroment]

    @property
    def stacks(self):
        if not Context.args.env:
            return self.Stacks

        return [stack for stack in self.Stacks if not stack.envs or Context.args.env in stack.envs]

    @staticmethod
    def parse(path: str = None):
        try:
            data = yaml.safe_load(open(path))
        except:
            raise IOError("Not able to find file at path: " + '"' + path + '"')
        try:
            return Config(**data)
        except Exception as e:
            raise Exception("Not able to pase config: " + '"' + str(e) + '"')

    def setup_env(self, env=None):
        _env = self.enviroment
        
        os.environ["AWS_PROFILE"] = _env.profile
        os.environ["AWS_DEFAULT_REGION"] = _env.region

    @property
    def enviroment(self):
        env = Context.args.env
        _env = None

        if env:
            for e in self.Environments:
                if e.name == env:
                    _env = e
                    break

        else:
            _env =  self.Environments[0]
        
        if not _env and env:
            raise Exception("Not able to find env with name " + env)
        
        if not _env:
            raise Exception("Not able to find env")

        return _env



from ..utils.logging import logger
from ..utils.config import Config
from ..utils.context import Context
import sys
import re
import json
from ..utils.common import create_change_set, remove_change_set, format_diff, get_yes_or_no, deploy_stack, create_stack,package


def deploy(config_path, root_path):
    config = Config.parse(config_path)
    config.setup_env(Context.get_args().env)
    services = config.stacks
    logger.warning(f"Checking difference for stacks from file {config_path}")
    
    logger.info(f"* Found {len(services)} services checking differences...")

    for service in services:
        if not re.search(Context.get_args().service, service.name):
            continue

        change_set = create_change_set(service, config)
        if change_set:
            diffs = [format_diff(change)for change in change_set["Changes"]]

            if len(diffs):
                logger.warning(f"Found {len(diffs)} differences for the stack {service.name}")
                for diff in diffs:
                    logger.warning(f"> {diff}")
                
                should_continue = get_yes_or_no(f"Do you wish to continue to update service: {service.name}")

                if not should_continue:
                    remove_change_set(service.name, change_set["ChangeSetName"])
                else:
                    logger.info("Deploying service...")
                    deploy_stack(service.name, change_set["ChangeSetName"])
            else:
                logger.info(f"Found no differences for the stack {service.name}")
        else:
            yml = package(service, config)
            logger.warn(f"{service.name} new stack ⭐")
            logger.warn(yml)
            should_continue = get_yes_or_no(f"Do you wish to continue to update service: {service.name}")

            if should_continue:
                create_stack(service, yml)

        



import argparse

from .meta import VERSION
from .commands.deploy import deploy
from .commands.init import init
from .commands.diff import diff
from .commands.info import info
from .commands.destroy import destroy
from .utils.logging import logger
from .utils.context import Context

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "action", 
        choices=['diff', 'info', 'deploy', 'package', "version", "init", "destroy"], 
        help='what action to preform'
    )

    parser.add_argument(
        "-s", "--service",
        default=".*",
        help="specify which stack"
    )

    parser.add_argument(
        "-path", "--path",
        default="services.yml",
        help='path to the file describing the services'
    )

    parser.add_argument(
        "-e", "--env",
        default="",
        help="which enviorment to use"
    )

    parser.add_argument("-v", '--version', action='version', help='path to the file describing the services', version='aws-cf: ' + VERSION)
    parser.add_argument("-r", "--root")
    parser.add_argument("-vb", "--verbose", type=bool)
    

    args = parser.parse_args()

    try:
        Context.set_root(args.root or ".")
        Context.set_args(args)
        Context.set_service_path(args.path)

        if args.action == "deploy":
            deploy(args.path, args.root or ".")

        if args.action == "destroy":
            destroy(args.path, args.root or ".")

        if args.action == "info":
            info()

        if args.action == "init":
            init()

        if args.action == "diff":
            diff(args.path, args.root or ".")

    except Exception as e:
        if args.verbose:
            raise e
        logger.error(str(e))



if __name__ == '__main__':
    main()


from ..utils.logging import logger
from ..utils.common import get_yes_or_no
import os 
import shutil

def init():
    logger.warn("Creating new aws-cf project")
    name = input("Enter name (default ./)")
    path = name or "./"
    create_aws_folder = get_yes_or_no("Do you want to create a aws folder in the project?")
    

    if os.path.exists(path) and len(os.listdir(f"{path}")) != 0:
        raise Exception("Init project needs to be in an empty directory")

    if create_aws_folder:
        source = os.path.dirname(os.path.realpath(__file__)) + "/../assets/default"
    else:
        source = os.path.dirname(os.path.realpath(__file__)) + "/../assets/simple"
        shutil.copytree(source, path + "/")
        


from typing import Any

class Context:
    root: str
    service_path: str
    args: Any

    @staticmethod
    def get_args():
        return Context.args

    @staticmethod
    def set_args(args):
        Context.args = args


    @staticmethod
    def get_changeset_prefix():
        return "AWSCF"

    @staticmethod
    def set_root(root: str):
        Context.root = root

    @staticmethod
    def get_root():
        return Context.root

    @staticmethod
    def set_service_path(path: str):
        Context.service_path = path

    @staticmethod
    def get_service_path():
        return Context.service_path


from setuptools import setup
from aws_cf.meta import VERSION

setup(
    name='aws-cf',
    version=VERSION,
    description='Simple way to deploy AWS stacks',
    long_description='Simple way to deploy AWS stacks',
    url='https://github.com/erikschmutz/aws-cf/',
    author='Erik Rehn',
    author_email='erik.rehn98@gmail.com',
    license='BSD 2-clause',
    packages=['aws_cf', 'aws_cf.utils', 'aws_cf.commands',
              'aws_cf.assets.simple', 'aws_cf.assets.default'],
    install_requires=['pydantic', 'boto3', 'pyaml'],
    entry_points={
        'console_scripts': [
            'aws-cf = aws_cf.__main__:main'
        ]
    },
    classifiers=[],
)



import logging

class Formatter(logging.Formatter):

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(message)s"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

logger = logging.getLogger("My_app")
logger.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(Formatter())

logger.addHandler(ch)






import boto3
import botocore.exceptions
import time
import subprocess
from .config import Config, Stack
from .context import Context
import tempfile
import json
import datetime

# 'StackStatus': 'CREATE_IN_PROGRESS'|'CREATE_FAILED'|'CREATE_COMPLETE'|'ROLLBACK_IN_PROGRESS'|'ROLLBACK_FAILED'|'ROLLBACK_COMPLETE'|'DELETE_IN_PROGRESS'|'DELETE_FAILED'|'DELETE_COMPLETE'|'UPDATE_IN_PROGRESS'|'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS'|'UPDATE_COMPLETE'|'UPDATE_FAILED'|'UPDATE_ROLLBACK_IN_PROGRESS'|'UPDATE_ROLLBACK_FAILED'|'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS'|'UPDATE_ROLLBACK_COMPLETE'|'REVIEW_IN_PROGRESS'|'IMPORT_IN_PROGRESS'|'IMPORT_COMPLETE'|'IMPORT_ROLLBACK_IN_PROGRESS'|'IMPORT_ROLLBACK_FAILED'|'IMPORT_ROLLBACK_COMPLETE',

def create_change_set(stack: Stack, config: Config):
    PREFIX = Context.get_changeset_prefix()

    root_path = Context.get_root()

    path = stack.path
    path = path.replace("$root", root_path)
    client = boto3.client("cloudformation")
    name = stack.name
    change_set_name = PREFIX + str(datetime.datetime.now().isoformat()).replace(":", "").split(".")[0]

    try:
        client.list_change_sets(StackName=stack.name)
    except botocore.exceptions.ClientError as e:
        if str(e).endswith(f"Stack [{stack.name}] does not exist"):
            return None
        raise e

    parameters  =[]
    
    if stack.parameters:
        parameters = [{"ParameterKey": key, "ParameterValue": stack.parameters[key]} for key in stack.parameters.keys()]

    client.create_change_set(
        ChangeSetName=change_set_name,
        StackName=name,
        Capabilities=["CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND"],
        TemplateBody=package(stack, config),
        Parameters=parameters,
    )
    wait_for_ready(name, change_set_name)

    return client.describe_change_set(
        ChangeSetName=change_set_name,
        StackName=name
    )


def wait_for_status(name: str, status):

    client = boto3.client("cloudformation") 

    iterations = 0
    MAX_ITERATIONS = 300
    SLEEP_SECONDS = 5
    while True:
        time.sleep(SLEEP_SECONDS)
        response = client.describe_stacks(
           StackName=name,
        )

        stack = response["Stacks"][0]

        if iterations > MAX_ITERATIONS:
            raise Exception(f"Stack {name} took more than {MAX_ITERATIONS*SLEEP_SECONDS} seconds to deploy.")

        if stack["StackStatus"] not in status:
            break

        iterations += 1


def wait_for_stack_deployed(name: str):
    return wait_for_status(name, ['CREATE_IN_PROGRESS', 'UPDATE_IN_PROGRESS'])

def wait_for_deleted(name: str):
    return wait_for_status(name, ['CREATE_IN_PROGRESS', 'UPDATE_IN_PROGRESS'])

def wait_for_ready(name, change_set_name):
    client = boto3.client("cloudformation") 
    while True:
        response = client.describe_change_set(
            ChangeSetName=change_set_name,
            StackName=name,
        )
    
        response = client.describe_change_set(
            ChangeSetName=change_set_name,
            StackName=name,
        )

        if response["Status"] not in ["CREATE_PENDING", "CREATE_IN_PROGRESS"]:
            break

        time.sleep(3)

def remove_change_set(name: str, change_set_name: str):
    client = boto3.client("cloudformation")

    response = client.delete_change_set(
        ChangeSetName=change_set_name,
        StackName=name
    )

def format_diff(diff):
    action = diff["ResourceChange"]["Action"]
    resource_id = diff["ResourceChange"]["LogicalResourceId"]
    resource_type = diff["ResourceChange"]["ResourceType"]
    details = diff["ResourceChange"]["Details"]

    actionName = {
        "Add": "Adding",
        "Modify": "Modifying",
        "Remove": "Removing"
    }

    if len(details):
        return f"{actionName[action]} {resource_type} with id {resource_id} \n{json.dumps(details)}\n\n"
        
    return f"{actionName[action]} {resource_type} with id {resource_id}"

def deploy_stack(name: str, change_set):
    client = boto3.client("cloudformation")
    response = client.execute_change_set(
        ChangeSetName=change_set,
        StackName=name
    )
    wait_for_stack_deployed(name)

def create_stack(stack: Stack, template):
    client = boto3.client("cloudformation")
    parameters  =[]
    
    if stack.parameters:
        parameters = [{"ParameterKey": key, "ParameterValue": stack.parameters[key]} for key in stack.parameters.keys()]

    response = client.create_stack(
        StackName=stack.name,
        TemplateBody=template, 
        Capabilities=["CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND"],
        Parameters=parameters
    )
    wait_for_stack_deployed(stack.name)

def delete_stack(name: str):
    client = boto3.client("cloudformation")
    client.delete_stack(
        StackName=name,
    )
    wait_for_deleted(name)

def package(stack: Stack, config: Config):
    args = [
            "aws", "cloudformation", "package",
            "--template", stack._path,
            "--s3-prefix", "aws/stacks",
            "--s3-bucket", config.enviroment.artifacts,
    ]
    
    if config.enviroment.profile:
        args.append("--profile")
        args.append(config.enviroment.profile)

    result = subprocess.check_output(args)
    return result.decode()

def get_yes_or_no(message):
    while True:
        result = input(message + " (enter y/n)")

        if result in ["yes", "y"]:
            return True

        if result in ["no", "n"]:
            return False